
CBV_CLASS_KEY = "__cbv_class__"
CLASS_TYPE = "__cbv_class__"

SIGNATURE_KEY = "__signature__"
API_METHODS = "__api_methods__"
INIT_MODIFIED = "__init_modified__"
GENERIC_ATTRIBUTES = "__generic_attribute__"
TAGGING_KEY = "TAGGING"